@@ -34,6 +34,7 @@
 #include <stdio.h>
 #include <errno.h>
 #include <string.h>
+#include <stdlib.h>
 #ifdef SYSSYSLOG
 #include <sys/syslog.h>
 #else
@@ -72,7 +73,12 @@ char *progname;
 /*************************************************************************/
 
 int
+#ifdef __STDC__
 parsetime(char *whattime)
+#else
+parsetime(whattime)
+char *whattime;
+#endif
 {
     static char *days[] =
     {"Su", "Mo", "Tu", "We", "Th", "Fr", "Sa", "Wk"};
@@ -95,27 +101,29 @@ parsetime(char *whattime)
     while (match && isalpha(*whattime) && isupper(*whattime)) {
         match = 0;
         for (loop = 0; loop < 8; loop++) {
-            if (strncmp(days[loop], whattime, 2) == NULL) {
+            if (strncmp(days[loop], whattime, 2) == 0) {
                 whattime += 2;
                 match = 1;
-                if ((wday == loop) | ((loop == 7) && wday && (wday < 6)))
+                if ((wday == loop) || ((loop == 7) && wday && (wday < 6)))
                     validday = 1;
             }
         }
     }
 
-    if (strncmp(whattime, "Any", 3) == NULL) {
-        validday = 1;
-        whattime += 3;
+    if (!validday) {
+        if (strncmp(whattime, "Any", 3) == 0) {
+            validday = 1;
+            whattime += 3;
+        }
+        else
+            return (0);
     }
-    if (!validday)
-        return (0);
 
     if (sscanf(whattime, "%d-%d", &start, &stop) == 2) {
         ltime = curtime->tm_min + 100 * curtime->tm_hour;
-        if ((start < stop) && ((ltime > start) && ltime < stop))
+        if ((start < stop) && ((ltime >= start) && ltime < stop))
             return (1);
-        if ((start > stop) && ((ltime > start) || ltime < stop))
+        if ((start > stop) && ((ltime >= start) || ltime < stop))
             return (1);
     } else
         return (1);
@@ -131,7 +139,12 @@ parsetime(char *whattime)
 /*************************************************************************/
 
 int
+#ifdef __STDC__
 validtime(char *ptr)
+#else
+validtime(ptr)
+char *ptr;
+#endif
 {
     char *nextptr;
     int good;
@@ -149,34 +162,48 @@ validtime(char *ptr)
     }
 }
 
+int
+#ifdef __STDC__
 acl_getlimit(char *aclbuf, char *class)
+#else
+acl_getlimit(aclbuf,class)
+char *aclbuf;
+char *class;
+#endif
 {
     char *crptr,
      *ptr,
       linebuf[1024];
     int limit;
 
-    while (*aclbuf != NULL) {
+    while (*aclbuf != '\0') {
         if (strncasecmp(aclbuf, "limit", 5) == 0) {
             for (crptr = aclbuf; *crptr++ != '\n';) ;
-            *--crptr = NULL;
+            *--crptr = '\0';
             strcpy(linebuf, aclbuf);
             *crptr = '\n';
             (void) strtok(linebuf, " \t");  /* returns "limit" */
-            if (strcmp(class, strtok(NULL, " \t")) == 0) {
-                limit = atoi(strtok(NULL, " \t"));  /* returns limit <n> */
-                if ((ptr = strtok(NULL, " \t")) && validtime(ptr))
-                    return (limit);
+            if ((ptr = strtok(NULL, " \t")) && (strcmp(class, ptr) == 0)) {
+                if ((ptr = strtok(NULL, " \t"))) {
+                    limit = atoi(ptr);  /* returns limit <n> */
+                    if ((ptr = strtok(NULL, " \t")) && validtime(ptr))
+                        return (limit);
+                }
             }
         }
         while (*aclbuf && *aclbuf++ != '\n') ;
     }
 
-    return (0);
-
+    return (-1);
 }
 
+int
+#ifdef __STDC__
 acl_countusers(char *class)
+#else
+acl_countusers(class)
+char *class;
+#endif
 {
     int pidfd,
       count,
@@ -206,14 +233,19 @@ struct flock arg;
             if (((stat == -1) && (errno == EPERM)) || !stat) {
                 if (strcmp(progname,"ftpcount")) {
 #if defined(SVR4)
-                    sprintf(line,"/bin/ps -l -p %d",buf[which]);
+#ifdef AIX
+                    sprintf(line,"/bin/ps %d",buf[which]);
+#elif defined(sun)
+                    sprintf(line,"/usr/ucb/ps auxww %d",buf[which]);
+#else
+                    sprintf(line,"/bin/ps -f -p %d",buf[which]);
+#endif
 #elif defined(M_UNIX)
                     sprintf(line,"/bin/ps -f -p %d",buf[which]);
 #else
                     sprintf(line,"/bin/ps %d",buf[which]);
 #endif
                     ZeFile = popen(line, "r");
-                    rewind(ZeFile);
                     fgets(line, 1024, ZeFile);
                     fgets(line, 1024, ZeFile);
                     printf("%s",line);
@@ -235,7 +267,12 @@ struct flock arg;
 }
 
 void
+#ifdef __STDC__
 new_list(struct c_list **list)
+#else
+new_list(list)
+struct c_list **list;
+#endif
 {
     (*list) = (struct c_list *) malloc(sizeof(struct c_list));
 
@@ -243,7 +280,13 @@ new_list(struct c_list **list)
 }
 
 int
+#ifdef __STDC__
 add_list(char *class, struct c_list **list)
+#else
+add_list(class,list)
+char *class;
+struct c_list **list;
+#endif
 {
     struct c_list *cp;
 
@@ -261,7 +304,14 @@ add_list(char *class, struct c_list **list)
     return (1);
 }
 
+int
+#ifdef __STDC__
 main(int argc, char **argv)
+#else
+main(argc,argv)
+int argc;
+char **argv;
+#endif
 {
     FILE *accessfile;
     char class[80],
@@ -294,20 +344,20 @@ main(int argc, char **argv)
         printf("%s: no service classes defined, no usage count kept\n",progname);
         exit(0);
     } else {
-        if (!(aclbuf = (char *) malloc(finfo.st_size + 1))) {
+        if (!(aclbuf = (char *) malloc((size_t)finfo.st_size + 1))) {
             perror("ftpcount: could not malloc aclbuf");
             exit(1);
         }
-        fread(aclbuf, finfo.st_size, 1, accessfile);
-        *(aclbuf + finfo.st_size) = '\0';
+        fread(aclbuf, (size_t)finfo.st_size, 1, accessfile);
+        *(aclbuf + (size_t)finfo.st_size) = '\0';
     }
 
     (void) new_list(&list);
     myaclbuf = aclbuf;
-    while (*myaclbuf != NULL) {
+    while (*myaclbuf != '\0') {
         if (strncasecmp(myaclbuf, "class", 5) == 0) {
             for (crptr = myaclbuf; *crptr++ != '\n';) ;
-            *--crptr = NULL;
+            *--crptr = '\0';
             strcpy(linebuf, myaclbuf);
             *crptr = '\n';
             (void) strtok(linebuf, " \t");  /* returns "class" */
@@ -318,16 +368,20 @@ main(int argc, char **argv)
                 ;
             } else {
                 limit = acl_getlimit(myaclbuf, class);
-            if (strcmp(progname,"ftpcount")) {
-                printf("Service class %s: \n", class);
-                printf("   - %3d users (%3d maximum)\n\n",
-            acl_countusers(class), limit);
-        }
-        else
-                printf("Service class %-20.20s - %3d users (%3d maximum)\n",
-                       class, acl_countusers(class), limit);
+                if (strcmp(progname,"ftpcount")) {
+                    printf("Service class %s: \n", class);
+                    printf("   - %3d users ", acl_countusers(class));
+                } else {
+                    printf("Service class %-20.20s - %3d users ",
+                            class, acl_countusers(class));
+                }
+                if (limit == -1)
+                    printf("(no maximum)\n");
+                else
+                    printf("(%3d maximum)\n", limit);
             }
         }
         while (*myaclbuf && *myaclbuf++ != '\n') ;
     }
+    exit(0);
 }
